#-------------------------------------------------------------------------------
# Name:        studio_utility
# Purpose:
## Author:      s.i.
#
# Created:     6/2015
# Copyright:   (c) MAS unit 2015
# Licence:     copy left
# Version:     1.0
#-------------------------------------------------------------------------------

import rhinoscriptsyntax as rs
import Rhino.Geometry as rg
import math
import random
import sys

bound_srf = None
bound_line = None
dumpmode = True

def DistanceSquared(p1,p2):
    v = rg.Vector3d(p1-p2)
    return math.pow(v.X,2) + math.pow(v.Y,2) + math.pow(v.X,2) 

def SetBoundary(site_surf):
    """
    site_surf: rhino surface object for boundary
    """
    global bound_srf
    global bound_line

    bound_srf = site_surf
    bound_line = rs.DuplicateSurfaceBorder(site_surf)

def getBoundMinMax():
    """
    get boundary 2d bounding box
       
    return:
        [rg.Point3d, rg.Point3d]
        minimum point maximum point
    """
    bbox = rs.BoundingBox(bound_srf)
    #return [rs.PointCoordinates(bbox[0]),rs.PointCoordinates(bbpx[6])]
    return [bbox[0],bbox[6]]

def dumpOff(mode = False):
    global dumpmode
    dumpmode = mode
    
def checkBrep(brep,breps):
    """
    brep: rhino brep object to check 
        if it is in boundary if it not collide with other 
        breps specified breps
    breps: hino brep objects to check 
    return:
        True: if it is in boundary, and not collide with others
        False: if it is not in boundary, or is colliedes with others
    """
    if bound_srf != None:
        bbitsc = rs.IntersectBreps(brep,bound_srf)
        if bbitsc == None:
            if dumpmode:
                print "is out range"
            return False
        bbitsc = rs.CurveBrepIntersect(bound_line,brep)
        if bbitsc != None:
            if dumpmode:
                print "is out range"
            return False
    for b in breps:
        bbitsc = rs.IntersectBreps(brep,b)
        if bbitsc != None:
            if dumpmode:
                print "is collisioned"
            return False
    return True        
    
def checkPoint(p,pts,buffer = 0):
    """
    p: rhino.Geometry point to check 
        if it is in boundary if it not closer with others than
        specified buffer
    pts: rhino.Geometry point list to check 
    buffer: buffer distance 
    return:
        True: if it is in boundary, and not closer than buffer with others
        False: if it is not in boundary, or is too close with others
    """
    if bound_srf != None:
        if rs.IsPointOnSurface(bound_srf,p) == False:
            if dumpmode:
                print "is out range"
            return False        
    for pt in pts:
        dist = rs.Distance(pt,p)
        if dist < buffer:
            if dumpmode:
                print "is too close"
            return False        
    return True

def rgPl2rsCV(pl):
    """
    generate Rhino polylineCurve from Rhino.Geometry polyline
    pl: rhino.Geometry polyline
    return:
        Rhino polylineCurve
    """
    pts= []
    for i in range(pl.PointCount):
        pts.append(pl.PointAt(i))
    return rs.AddPolyline(pts)

def getCenterPoint(pl):
    """
    pl: rhino.Geometry polyline to check 
    return:
        rg.Point3d
    """
    cp = rg.Point3d(0,0,0)
    for i in range(pl.PointCount):
        cp += pl.PointAt(i)
    if pl.PointCount > 0:
        cp /= pl.PointCount
    return cp

def checkPolyline(pl,pls,mode = 0):
    """
    pl: rhino.Geometry polyline to check 
        if it is in boundary if it not closer with others than
        specified buffer
    pls: rhino.Geometry polyline list as others
    mode: 
        0: no colide, no touch
        1: allow touch at line, touch at point
        2: allow touch at line, but not touch at point
        3: allow touch at a point, but not touch at line
    return:
        True: if it is in boundary, and does not colide others
        False: if it is not in boundary, or colides others
    """
    
    cv = None
    cv2 = None
    sf = None
    sf2 = None
    if bound_srf != None:
        cv = rgPl2rsCV(pl)
        it = rs.CurveSurfaceIntersection(cv,bound_srf)
        if it == None:
            if dumpmode:
                print "is out range"
            if sf != None:
                rs.DeleteObject(sf)
            if cv != None:
                rs.DeleteObject(cv)
            return False 
        it = rs.CurveCurveIntersection(bound_line,cv)
        if it != None:
            if dumpmode:
                print "is out range"
            if sf != None:
                rs.DeleteObject(sf)
            if cv != None:
                rs.DeleteObject(cv)
            return False
            
    if cv == None:
        cv = rgPl2rsCV(pl)
    sf = rs.AddPlanarSrf(cv)    
    cp = getCenterPoint(pl)

    if rs.IsPointOnSurface(bound_srf,cp) == False:
        if sf != None:
            rs.DeleteObject(sf)
        if cv != None:
            rs.DeleteObject(cv)
        return False
        
    for ppl in pls:
        cv2 = rgPl2rsCV(ppl)
        sf2 = rs.AddPlanarSrf(cv2)
        it = rs.IntersectBreps(sf,sf2)        
        if mode == 0:
            #line NG
            #point NG
            if it != None:    
                if sf != None:
                    rs.DeleteObject(sf)
                if cv != None:
                    rs.DeleteObject(cv)
                if sf2 != None:
                    rs.DeleteObject(sf2)
                if cv2 != None:
                    rs.DeleteObject(cv2)
                return False
            else:
                for p in rs.CurvePoints(cv):
                    for pp in rs.CurvePoints(cv2):
                        if rs.Distance(p,pp) < 0.1:
                            if dumpmode:
                                print "touch at a point"
                            if sf != None:
                                rs.DeleteObject(sf)
                            if cv != None:
                                rs.DeleteObject(cv)
                            if sf2 != None:
                                rs.DeleteObject(sf2)
                            if cv2 != None:
                                rs.DeleteObject(cv2)
                            return False
        elif mode == 1:
            #line ok
            #point ok
            if it != None:
                if rs.IsLine(it[0]) == False:
                    if dumpmode:
                        print "colide"
                        if sf != None:
                            rs.DeleteObject(sf)
                        if cv != None:
                            rs.DeleteObject(cv)
                        if sf2 != None:
                            rs.DeleteObject(sf2)
                        if cv2 != None:
                            rs.DeleteObject(cv2)
                    return False                    
        elif mode == 2:
            #line ok
            #point NG
            if it != None:
                if rs.IsLine(it[0]) == False:
                    if dumpmode:
                        print "colide"
                    if sf != None:
                        rs.DeleteObject(sf)
                    if cv != None:
                        rs.DeleteObject(cv)
                    if sf2 != None:
                        rs.DeleteObject(sf2)
                    if cv2 != None:
                        rs.DeleteObject(cv2)
                    return False
            else:
                for p in rs.CurvePoints(cv):
                    for pp in rs.CurvePoints(cv2):
                        if rs.Distance(p,pp) < 0.1:
                            if dumpmode:
                                print "touch at a point"
                            if sf != None:
                                rs.DeleteObject(sf)
                            if cv != None:
                                rs.DeleteObject(cv)
                            if sf2 != None:
                                rs.DeleteObject(sf2)
                            if cv2 != None:
                                rs.DeleteObject(cv2)
                            return False
        elif mode == 3:
            #line NG
            #point OK
            if it != None:
                if dumpmode:
                    print "colide"
                if sf != None:
                    rs.DeleteObject(sf)
                if cv != None:
                    rs.DeleteObject(cv)
                if sf2 != None:
                    rs.DeleteObject(sf2)
                if cv2 != None:
                    rs.DeleteObject(cv2)
                return False
        
    if sf != None:
        rs.DeleteObject(sf)
    if cv != None:
        rs.DeleteObject(cv)
    if sf2 != None:
        rs.DeleteObject(sf2)
    if cv2 != None:
        rs.DeleteObject(cv2)
    
    return True
    
def w_rand_choice(units):
    """
    units: list of instance
           1 instance should have member 
           .w
           which express weight                      
    return:
        1 instance from list "units"
        selected randomly weighted probability of ".w"
    """
    n = None
    while n == None:
        s = 0
        for u in units:
            s += u.w            
            if random.random()*s < u.w:
                n = u
    return n

class _Unit:
    def __init__(self,pos):
        self.pt = None
        self.pos = pos
        self.group = 0
        self.w = 0
        
#k-means
def k_means(units, clus): 
    """
    units: list of instance
           1 instance should have 2 member 
           .group (integer)
           .pos (Rhino.Geometry.Point3d)                  
    clus: expected cluster number (integer)
        
    return:
        list of [rg.Point3d, num]
        [0]center point of cluster
        [1]cluster group id
    """
    random.seed(None) 
    for u in units:
        u.group = random.randint(0,clus-1)
    
    wps = []
    for i in range(clus):
        wps.append([rg.Point3d(0,0,0),0])
    
    tryal = 0
    while 1:
        #reset
        for wp in wps:
            wp[0] = rg.Point3d(0,0,0)
            wp[1] = 0
            
        for u in units:
            wps[u.group][0] += u.pos
            wps[u.group][1] += 1
        
        for wp in wps:
            if wp[1] > 0:
                wp[0] /= wp[1]
                    
        revised = False
        for u in units:        
            dist = rs.Distance(u.pos,wps[u.group][0])
            for i in range(len(wps)):
                dist2 = rs.Distance(u.pos,wps[i][0])
                if dist2 < dist:
                    u.group = i
                    revised = True
        if revised == False:
            print "tryal",tryal
            break
        tryal+=1
        
    for i in range(len(wps)):
        wps[i][1] = i
    return wps
    
#k-means++
def k_meansPP(units, clus):
    """
    units: list of instance
           1 instance should have 3 member 
           .group (integer)
           .pos (Rhino.Geometry.Point3d)      
           .w (double)
    clus: expected cluster number (integer)
    
    return:
        list of [rg.Point3d, num]
        [0]center point of cluster
        [1]cluster group id
    """
    random.seed(None)
    wps = []
    wu = units[random.randint(0,len(units)-1)]
    wps.append([wu.pos,0])
    
    candinates = list(units)
    for i in range(clus-1):
        candinates.remove(wu)
        sum_pow_d = 0
        for u in units:
            closest = sys.float_info.max
            closestwp = None
            for wp in wps:
                distsq = DistanceSquared(u.pos,wp[0])
                if closest > distsq:
                    closest = distsq
                    closestwp = wp
            pow_d = DistanceSquared(closestwp[0],u.pos)
            u.w = pow_d
            sum_pow_d += pow_d        
        for u in units:
            u.w = u.w / sum_pow_d     
        wu = w_rand_choice(candinates)
        wps.append([wu.pos,0])
        
    for u in units:
        closest = sys.float_info.max
        closestid = 0
        for i in range(len(wps)):
            distsq = DistanceSquared(wps[i][0],u.pos)
            if closest > distsq:
                closest = distsq
                closestid = i
        u.group = closestid
    
    tryal = 0
    while 1:
        units_bf = list(units)
        #reset
        for wp in wps:
            if tryal!=0:
                wp[0] = rg.Point3d(0,0,0)
            wp[1] = 0
        for u in units:
            wps[u.group][0] += u.pos
            wps[u.group][1] += 1
        for wp in wps:
            if wp[1] > 0:
                wp[0] /= wp[1]
        revised = False
        for u in units:
            distsq = DistanceSquared(u.pos,wps[u.group][0])
            for i in range(len(wps)):
                distsq2 = DistanceSquared(u.pos,wps[i][0])
                if distsq2 < distsq:
                    u.group = i
                    revised = True
        if revised == False:
            print "tryal",tryal
            break
        tryal+=1
    for i in range(len(wps)):
        wps[i][1] = i
    return wps

class Cluster:
    def __init__(self,id,u):
        self.id = id
        self.cp = rg.Point3d(0,0,0)
        self.us = set()
        self.add(u)
        self.sumdsq = 0
    def hasunit(self,u):
        for uu in self.us:
            if u == uu:
                return True
        return False
    def idShareUnit(self,clu):
        for u in clu.us:
            for uu in self.us:
                if u == uu:
                    return True
        return False
    def add(self,u):
        u.group = self.id
        self.us.add(u)
        self.cp = rg.Point3d(0,0,0)
        for u in self.us:
            self.cp += u.pos
        if len(self.us)>1:
            self.cp /= len(self.us)
        self.sumdsq = self.sumDistSq(self.us)
    def sumDistSq(self,list):
        cp = rg.Point3d(0,0,0)
        for u in list:
            cp += u.pos
        if len(list) > 1:
            cp /= len(list)
        sumdsq = 0
        for u in list:
            sumdsq += DistanceSquared(u.pos,cp)
        return sumdsq
    def merge(self,clu):
        for u in clu.us:
            self.add(u)
        clu.us = set()
    def mergeTest(self,clu):        
        templist = list(list(self.us) + list(clu.us))
        return self.sumDistSq(templist)
    def label(self,id):
        for u in self.us:
            u.group = id


def wardMethodStep(units,steps):
    """
    units: list of instance
           1 instance should have 2 member 
           .group (integer)
           .pos (Rhino.Geometry.Point3d)      
    steps: clustering steps (integer)
    
    return:
        list of [rg.Point3d, num]
        [0]center point of cluster
        [1]cluster group id
    """    
    clusters = []
    for u in units:
        u.group = -1
    
    prs =[]
    for u in units:
        closest = sys.float_info.max
        uc = None
        for uu in units:
            if u != uu:
                distsq = DistanceSquared(u.pos,uu.pos)
                if closest > distsq:
                    closest = distsq
                    uc = uu
        pr = [u,uc]
        prs.append(pr)
        isClulstered = False
        for cu in clusters:
            if cu.hasunit(u):
                cu.add(uc)
                isClulstered = True
                break
            if cu.hasunit(uc):
                cu.add(u)
                isClulstered = True
                break
        if isClulstered == False:
            cu = Cluster(len(clusters),u)
            cu.add(uc)
            clusters.append(cu)
            
        for cu in clusters:
            for cuu in clusters:
                if cu != cuu:
                    if cu.idShareUnit(cuu):
                        cu.merge(cuu)
    
    for cu in clusters:
        if len(cu.us) == 0:
            clusters.remove(cu)
    
    for j in range(steps):
        shortest = sys.float_info.max
        mergePair = []
        for i in range(len(clusters)-1):
            cu = clusters[i]
            for ii in range(len(clusters)):
                cuu = clusters[ii]
                if cu != cuu:
                    sumdsq = cu.mergeTest(cuu)
                    sumdsqdiff = sumdsq - cu.sumdsq - cuu.sumdsq
                    if shortest > sumdsqdiff:
                        shortest = sumdsqdiff
                        mergePair = [cu,cuu]
        
        if len(mergePair) > 0:
            mergePair[0].merge(mergePair[1])
        
        for cu in clusters:
            if len(cu.us) == 0:
                clusters.remove(cu)
        
    for i in range(len(clusters)):
        clusters[i].label(i)
    
    wps = []
    for cl in clusters:
        wp = [cl.cp,cl.id]
        wps.append(wp)
    return wps

def wardMethodCNum(units,clus):
    """
    units: list of instance
           1 instance should have 2 member 
           .group (integer)
           .pos (Rhino.Geometry.Point3d)      
    clus: cluster number (integer)
    
    return:
        list of [rg.Point3d, num]
        [0]center point of cluster
        [1]cluster group id
    """    
    clusters = []
    for u in units:
        u.group = -1
    
    prs =[]
    for u in units:
        closest = sys.float_info.max
        uc = None
        for uu in units:
            if u != uu:
                distsq = DistanceSquared(u.pos,uu.pos)
                if closest > distsq:
                    closest = distsq
                    uc = uu
        pr = [u,uc]
        prs.append(pr)
        isClulstered = False
        for cu in clusters:
            if cu.hasunit(u):
                cu.add(uc)
                isClulstered = True
                break
            if cu.hasunit(uc):
                cu.add(u)
                isClulstered = True
                break
        if isClulstered == False:
            cu = Cluster(len(clusters),u)
            cu.add(uc)
            clusters.append(cu)
            
        for cu in clusters:
            for cuu in clusters:
                if cu != cuu:
                    if cu.idShareUnit(cuu):
                        cu.merge(cuu)
    
    for cu in clusters:
        if len(cu.us) == 0:
            clusters.remove(cu)
    
    while 1:
        shortest = sys.float_info.max
        mergePair = []
        for i in range(len(clusters)-1):
            cu = clusters[i]
            for ii in range(len(clusters)):
                cuu = clusters[ii]
                if cu != cuu:
                    sumdsq = cu.mergeTest(cuu)
                    sumdsqdiff = sumdsq - cu.sumdsq - cuu.sumdsq
                    if shortest > sumdsqdiff:
                        shortest = sumdsqdiff
                        mergePair = [cu,cuu]
        
        if len(mergePair) > 0:
            mergePair[0].merge(mergePair[1])
        else:
            break
        for cu in clusters:
            if len(cu.us) == 0:
                clusters.remove(cu)        
        if len(clusters) == clus:
            break
        
    for i in range(len(clusters)):
        clusters[i].label(i)
    
    wps = []
    for cl in clusters:
        wp = [cl.cp,cl.id]
        wps.append(wp)
    return wps

#k-means rhino Geometry point
def k_meansRGP(pts,clus):
    """
    clustering by k-means
    pts: list of Rhino.Geometry.Point3d
    clus: expected cluster number (integer)
    
    return:
        list of clusters [list of Point3d]
    """
    units = []
    for pt in pts:
        u = _Unit(rs.PointCoordinates(pt))
        u.pt = pt
        units.append(u)
    k_means(units, clus)
    units = sorted(units, key = lambda x: x.group)
    ptss = []
    group = 0
    for u in units:
        if len(ptss)-1 != u.group:
            ptss.append(list())
        ptss[u.group].append(u.pt)
    return ptss

#k-means++ rhino Geometry point
def k_meansPPRGP(pts,clus):
    """
    clustering by k-means++
    pts: list of Rhino.Geometry.Point3d
    clus: expected cluster number (integer)
    
    return:
        list of clusters [list of Point3d]
    """
    units = []
    for pt in pts:
        u = _Unit(rs.PointCoordinates(pt))
        u.pt = pt
        units.append(u)
    k_meansPP(units, clus)
    units = sorted(units, key = lambda x: x.group)
    ptss = []
    group = 0
    for u in units:
        if len(ptss)-1 != u.group:
            ptss.append(list())
        ptss[u.group].append(u.pt)
    return ptss

#ward method step Geometry point
def wardMethodStepRGP(pts,steps):
    """
    clustering by ward method
    pts: list of Rhino.Geometry.Point3d
    steps: number of steps (integer)
    
    return:
        list of clusters [list of Point3d]
    """
    units = []
    for pt in pts:
        u = _Unit(rs.PointCoordinates(pt))
        u.pt = pt
        units.append(u)
    wardMethodStep(units, steps)
    units = sorted(units, key = lambda x: x.group)
    ptss = []
    group = 0
    for u in units:
        if len(ptss)-1 != u.group:
            ptss.append(list())
        ptss[u.group].append(u.pt)
    return ptss

#ward method cnum Geometry point
def wardMethodCNumRGP(pts,clus):
    """
    clustering by ward method
    pts: list of Rhino.Geometry.Point3d
    clus: expected cluster number (integer)
    
    return:
        list of clusters [list of Point3d]
    """
    units = []
    for pt in pts:
        u = _Unit(rs.PointCoordinates(pt))
        u.pt = pt
        units.append(u)
    wardMethodCNum(units, clus)
    units = sorted(units, key = lambda x: x.group)
    ptss = []
    group = 0
    for u in units:
        if len(ptss)-1 != u.group:
            ptss.append(list())
        ptss[u.group].append(u.pt)
    return ptss

def explodeRsPolyline(pl, delete_input = False):
    """
    explode Polyline
    pl: PolyLine
    delete_input: 
            True: delete original polyline
            False: keep original polyline
            (option)
    
    return:
        list of Line
    """
    llist = []
    plist = rs.PolylineVertices(pl)
    for i in range(len(plist)-1):
        l = rs.AddLine(plist[i],plist[i+1])
        llist.append(l)
    if rs.IsCurveClosed(pl):
        l = rs.AddLine(plist[0],plist[len(plist)-1])
        llist.append(l)
    if delete_input:
        rs.DeleteObject(pl)
    return llist
    
def delDupLineInList(lineList, precision = 0.1):
    """
    delete duplication, lineList must be purely line list (do not contain other type objects
    lineList: list of line
    precision: 
            precision of point(start / end) matching 
            (option)
    
    return:
        it returns nothing, original lineList would be modified
    """
    duplist = []
    for i in range(len(lineList)-1):
        l = lineList[i]
        sp1 = rs.CurveStartPoint(l)
        ep1 = rs.CurveEndPoint(l)
        for ii in range(i + 1, len(lineList)):
            ll = lineList[ii]
            sp2 = rs.CurveStartPoint(ll)
            ep2 = rs.CurveEndPoint(ll)
            if rs.Distance(sp1,sp2) < precision and rs.Distance(ep1,ep2) < precision:
                duplist.append(ll)
            elif rs.Distance(sp1,ep2) < precision and rs.Distance(ep1,sp2) < precision:
                duplist.append(ll)
    for dl in duplist:
        lineList.remove(dl)
    
    if dumpmode:
        print "dup ",len(duplist)

def extendLineListWODup(llist1, llist2, precision = 0.1):
    """
    extend 2 line Lists without dulplication, be purely line list (do not contain other type objects)
    llist1: line to add
    llist2: line list to extend
    precision: 
            precision of point(start / end) matching 
            (option)
    
    return:
        returns extended line list
    """
    result = list(llist2)
    for l in llist1:
        sp1 = rs.CurveStartPoint(l)
        ep1 = rs.CurveEndPoint(l)
        isNotInList = True
        for ll in llist2:
            sp2 = rs.CurveStartPoint(ll)
            ep2 = rs.CurveEndPoint(ll)
            if rs.Distance(sp1,sp2) < precision and rs.Distance(ep1,ep2) < precision:
                isNotInList = False
                break
            elif rs.Distance(sp1,ep2) < precision and rs.Distance(ep1,sp2) < precision:
                isNotInList = False
                break
        if isNotInList:
            result.append(l)
    return result
    
def convertRSLinesToRGLines(rslines):
    """
    cnvert lines 
    rslines: line list to convert (lines from rhinoscriptsyntax)
    
    return:
        returns list of Rhinog.Geometry.line
    """
    result = []
    for l in rslines:
        sp = rs.CurveStartPoint(l)
        ep = rs.CurveEndPoint(l)
        rl = rg.Line(sp,ep)
        result.append(rl)
    return result
    
def convertRGLinesToRSLines(rglines):
    """
    cnvert lines 
    rglines: line list to convert (Rhinog.Geometry.line)
    
    return:
        returns list of line (from rhinoscriptsyntax)
    """
    result = []
    for l in rglines:
        rl = rs.AddLine(l.From,l.To)
        result.append(rl)
    return result

def roundRg3dPt(pt,position):
    """
    round of Rhino.Geometry.Point3D coordinate 
    pt: 
        Rhino.Geometry.Point3D
    position:
        round position
        
    return:
        Rhino.Geometry.Point3D
    """
    npt = rg.Point3d(pt)
    npt.X = round(pt.X,position)
    npt.Y = round(pt.Y,position)
    npt.Z = round(pt.Z,position)
    return npt
    
def roundLineCVs(cvs,position):
    """
    round of Line or Polyline  start and end point coordinate 
    cvs: 
        list of curve (Line or Polyline)
    position:
        round position
        
    return:
        rounded curve list
    """
    result = []
    for cv in cvs:
        if rs.IsLine(cv):
            sp = rs.CurveStartPoint(cv)
            ep = rs.CurveEndPoint(cv)
            l = rs.AddLine(roundRg3dPt(sp,position),roundRg3dPt(ep,position))
            result.append(l)
        if rs.IsPolyline(cv):
            vts= rs.PolylineVertices(cv)
            nvts = []
            for i in range(len(vts)):
                if i == 0:
                    sp = roundRg3dPt(vts[i],position)
                    nvts.append(sp)
                elif i == len(vts) - 1:
                    ep = roundRg3dPt(vts[i],position)
                    nvts.append(ep)
                else:
                    nvts.append(vts[i])
            pl = rs.AddPolyline(nvts)
            result.append(pl)            
    return result
    
def weighted_choice(ls):
    """
    weighted random choice
    ls: 
        list should be [['a',0.5],['b',3.5], ......]
        
    return:
        value at index 0 of list of list
    """
    x=random.random() * sum(t[1] for t in ls)
    for k, w in ls:
        x-=w
        if x<0: 
            return k