import graph_class_1_2 as graph
import rhinoscriptsyntax as rs
import scriptcontext as sc
import Rhino.Geometry as rg
import ghpythonlib.components as ghp
import sys

def splitCell(cell,lines):
    srf = rs.AddPlanarSrf([cell])
    clines = []
    for l in lines:
        clines.append(rs.coercecurve(l))
    srfs = ghp.SurfaceSplit(rs.coercebrep(srf),clines)
    #print type(srfs) == list
    if srfs == None:
        return None
    if type(srfs) == list:
        if len(srfs) == 2:
            sp = rs.CurveStartPoint(lines[0])
            cvs = []
            v1 = rg.Vector3d(rs.CurveEndPoint(lines[0]) - rs.CurveStartPoint(lines[0]))
            left = None
            right = None
            for srf in srfs:
                pl = rs.DuplicateSurfaceBorder(srf)
                cp = rs.SurfaceAreaCentroid(srf)[0]
                pm = rs.CurveClosestPoint(lines[0],cp)
                clp = rs.EvaluateCurve(lines[0],pm)
                vec = rg.Vector3d(cp-clp)        
                if rs.VectorCrossProduct(v1,vec)[2] > 0:
                    left = pl[0]
                else:
                    right = pl[0]
            return [left,right]
        else:
            cvs = []
            for srf in srfs:
                pl = rs.DuplicateSurfaceBorder(srf)
                cvs.extend(pl)
            return cvs
    else:
        pl = rs.DuplicateSurfaceBorder(srfs)
        return [pl]

def generateCells(divunit,bound_r,graph):
    """
    generate cells onto your graph(to nodes, to edges)
    divunit: division width(of a cell)
    bound_r: boundary radius
    graph: graph instance from module "graph_class_1_2"
    
    return:
        Nothing,
        graph will have new property
        Node.cell (closed curve) undivided
        Node.cell_devided (list of closed curve)
        Edge.cells (list of closed curve) undivided
        Edge.cells_left (list of closed curve)
        Edge.cells_right (list of closed curve)
        Edge.cells_right (list of closed curve)
    """
    #add new property
    for e in graph.edges:
        e.cellpts = []
        e.cells = []
    
    npts = []
    npts_flat = []
    #add new property
    for n in graph.nodes:
        npts.append(n.pos)
        npts_flat.append(rg.Point3d(n.pos.X,n.pos.Y,0.0))
        n.cell = None
        n.cell_flat = None
       
    epts = []
    epts_flat = []
    for e in graph.edges:
        segments = int(e.len0 / divunit)
        res = rs.DivideCurve(e.cv,segments)
        if res != None:
            for pt in res:
                if pt != e.n1.pos and pt != e.n2.pos:
                    epts.append(pt)
                    epts_flat.append(rg.Point3d(pt.X,pt.Y,0.0))
                    e.cellpts.append(pt)
        
    pts = []
    pts.extend(npts)
    pts.extend(epts)
    pts_flat = []
    pts_flat.extend(npts_flat)
    pts_flat.extend(epts_flat)
    #gen voronoi cell
    #cells = ghp.Voronoi(pts,bound_r)
    cells = ghp.Voronoi(pts_flat,bound_r)
    
    #creat dictionaly to attach cells to nodes and edges as property
    cellsmap = {}
    for i in range(len(cells)):
        pair = {pts[i]:cells[i]}
        cellsmap.update(pair)
    
    for n in graph.nodes:
        cell = cellsmap[n.pos]
        mcell = ghp.Move(cell,rg.Vector3d(0,0,n.pos.Z))
        #mcell = [cell]
        n.cell = mcell[0]
    
    for e in graph.edges:
        for pt in e.cellpts:
            cell = cellsmap[pt]
            mcell = ghp.Move(cell,rg.Vector3d(0,0,pt.Z))
            #mcell = [cell]
            e.cells.append(mcell[0])
            
    #split cell
    for e in graph.edges:
        #add new property
        e.cells_left = []
        e.cells_right = []
        for cell in e.cells:
            if rs.IsCurveClosed(cell) and rs.IsCurve(e.cv):
                cellpair = splitCell(cell,[e.cv])    
                if cellpair != None and len(cellpair) == 2:
                    e.cells_left.append(cellpair[0])
                    e.cells_right.append(cellpair[1])
                    
    #split cell at node
    for n in graph.nodes:
        if len(n.edges) > 1:
            ls = []
            for e in n.edges:
                ls.append(e.cv)
            n.cell_divided = splitCell(n.cell,ls)
        else:
            n.cell_divided = [n.cell]

def generateCellsWL(divlistlist,bound_r,graph):
    """
    generate cells onto your graph(to nodes, to edges)
    divlistlist: list of division width list [[width1,width2,width3,],[width1,width2,,,],,]
    bound_r: boundary radius
    graph: graph instance from module "graph_class_1_2"
    
    return:
        Nothing,
        graph will have new property
        Node.cell (closed curve) undivided
        Node.cell_devided (list of closed curve)
        Edge.cells (list of closed curve) undivided
        Edge.cells_left (list of closed curve)
        Edge.cells_right (list of closed curve)
        Edge.cells_right (list of closed curve)
    """
    #add new property
    for e in graph.edges:
        e.cellpts = []
        e.cells = []
    
    npts = []
    npts_flat = []
    #add new property
    for n in graph.nodes:
        npts.append(n.pos)
        npts_flat.append(rg.Point3d(n.pos.X,n.pos.Y,0.0))
        n.cell = None
        n.cell_flat = None
       
    epts = []
    epts_flat = []
    for i in range(min(len(graph.edges),len(divlistlist))):
        e = graph.edges[i]
        dlist = divlistlist[i]
        cv = rs.CopyObject(e.cv)
        for ii in range(len(dlist)):
            width = dlist[ii]
            pms = rs.DivideCurveEquidistant(cv,width,False,False)     
            pt = rs.EvaluateCurve(cv,pms[1])
            cv = rs.SplitCurve(cv,pms[1])[1]
            epts.append(pt)
            epts_flat.append(rg.Point3d(pt.X,pt.Y,0.0))
            e.cellpts.append(pt)

#           
#        
#        segments = int(e.len0 / divunit)
#        res = rs.DivideCurve(e.cv,segments)
#        if res != None:
#            for pt in res:
#                if pt != e.n1.pos and pt != e.n2.pos:
#                    epts.append(pt)
#                    epts_flat.append(rg.Point3d(pt.X,pt.Y,0.0))
#                    e.cellpts.append(pt)
        
    pts = []
    pts.extend(npts)
    pts.extend(epts)
    pts_flat = []
    pts_flat.extend(npts_flat)
    pts_flat.extend(epts_flat)

    
    #gen voronoi cell
    #cells = ghp.Voronoi(pts,bound_r)
    cells = ghp.Voronoi(pts_flat,bound_r)
    #objs.extend(cells)
    #print cells
    #creat dictionaly to attach cells to nodes and edges as property
    cellsmap = {}
    for i in range(len(cells)):
        pair = {pts[i]:cells[i]}
        cellsmap.update(pair)
    
    for n in graph.nodes:
        cell = cellsmap[n.pos]
        mcell = ghp.Move(cell,rg.Vector3d(0,0,n.pos.Z))
        #mcell = [cell]
        n.cell = mcell[0]
    
    for e in graph.edges:
        for pt in e.cellpts:
            cell = cellsmap[pt]
            mcell = ghp.Move(cell,rg.Vector3d(0,0,pt.Z))
            #mcell = [cell]
            e.cells.append(mcell[0])
            
    #split cell
    for e in graph.edges:
        #add new property
        e.cells_left = []
        e.cells_right = []
        for cell in e.cells:
            if rs.IsCurveClosed(cell) and rs.IsCurve(e.cv):
                cellpair = splitCell(cell,[e.cv])    
                if cellpair != None and len(cellpair) == 2:
                    e.cells_left.append(cellpair[0])
                    e.cells_right.append(cellpair[1])
                    
    #split cell at node
    for n in graph.nodes:
        if len(n.edges) > 1:
            ls = []
            for e in n.edges:
                ls.append(e.cv)
            n.cell_divided = splitCell(n.cell,ls)
        else:
            n.cell_divided = [n.cell]
            
def isAllCellsFitIn(boundary,graph):
    """
    check if all cells (belong to graph.edges, graph.nodes)
    are in the boundary on planar dimension
    
    return:
        True: if all cells are in
        False: if cell(s) is out of boundary, is intersect the pelimeter
    """
    
    fboundary = rs.CopyObject(boundary)
    cp = rs.SurfaceAreaCentroid(fboundary)[0]
    fboundary = rs.MoveObject(fboundary,[0,0,-cp.Z])
    plane = rg.Plane(rg.Point3d(0,0,0),rg.Vector3d(0,0,1))
    
    cells = []
    isFitIn = True
    for e in graph.edges:
        if isFitIn:
            for cell in e.cells_left:
                sf = rs.AddPlanarSrf([cell])
                cp = rs.SurfaceAreaCentroid(sf)[0]
                sf = rs.MoveObject(sf,[0,0,-cp.Z]) 
                res = rs.IntersectBreps(sf,fboundary)
                if res != None and len(res):
                    area1 = rs.CurveArea(cell)[0]
                    area2 = rs.CurveArea(res[0])[0]
                    if abs(area1 - area2) > 1:
                        print "not fit in", area1 - area2
                        isFitIn = False
                        break
                else:
                    print "not fit in None",res
                    isFitIn = False
                    break
        if isFitIn:
            for cell in e.cells_right:
                sf = rs.AddPlanarSrf([cell])
                cp = rs.SurfaceAreaCentroid(sf)[0]
                sf = rs.MoveObject(sf,[0,0,-cp.Z]) 
                res = rs.IntersectBreps(sf,fboundary)
                if res != None and len(res):
                    area1 = rs.CurveArea(cell)[0]
                    area2 = rs.CurveArea(res[0])[0]
                    if abs(area1 - area2) > 1:
                        print "not fit in", area1 - area2
                        isFitIn = False
                        break
                else:
                    print "not fit in None",res
                    isFitIn = False
                    break
        for n in graph.nodes:
            if isFitIn:
                for cell in n.cell_divided:
                    sf = rs.AddPlanarSrf([cell])
                    cp = rs.SurfaceAreaCentroid(sf)[0]
                    sf = rs.MoveObject(sf,[0,0,-cp.Z]) 
                    res = rs.IntersectBreps(sf,fboundary)
                    if res != None and len(res):
                        area1 = rs.CurveArea(cell)[0]
                        area2 = rs.CurveArea(res[0])[0]
                        if abs(area1 - area2) > 1:
                            print "not fit in", area1 - area2
                            isFitIn = False
                            break
                    else:
                        print rs.CurveArea(cell)
                        print "not fit in None",res,cell,sf
                        #objs.append(sf)
                        objs.append(cell)
                        isFitIn = False
                        break
    return isFitIn